@@ -64,7 +64,7 @@ on_window1_destroy                     (GtkObject       *object,
                                         gpointer         user_data)
 {
 	/* this must be done before calling SPI_event_quit */
-	gok_spy_deregister_mousebuttonlistener ((void *)gok_main_button_listener);
+	gok_spy_deregister_mousebuttonlistener ((void *)gok_main_mouse_button_listener);
 
 	SPI_event_quit ();
 }
